package tpl2java;

import java.util.Scanner;

/**
 * @author damien 
 * Classe de calcul sur des polynomes à coefficients réels.
 *
 */

/**
 * @author damien
 *
 */
public class Polynomes {
	/**
	 * Le tableau de coefficients
	 */
	private double monomes[];

	/**
	 * Le degré du polynome
	 */
	private int degre;
	


	// --------------------------------------------------------------------------------------------------------------
	
	/**
	 * Constructeur par défaut, on crée un polynome "null"
	 */
	public Polynomes() {
		super();
		degre = 0;
		monomes = new double[1];
	}

	/**
	 * Constructeur on passe le degré du polynome, cela crée néanmoins un polynome
	 * nul.
	 * 
	 * @param deg degré du polynome à créer
	 */
	public Polynomes(int deg) {
		super();
		if (deg < 0) {
			throw new IllegalArgumentException("Pas d'exposant négatif ! " + deg);
		}
		degre = deg;
		monomes = new double[degre + 1];
	}

	/**
	 * Constructeur recopiant un polynome
	 * 
	 * @param p polynome à recopier
	 */
	public Polynomes(Polynomes p) {
		super();
		degre = p.degre;
		monomes = new double[degre + 1];
		for (int i = 0; i <= degre; i++)
			monomes[i] = p.monomes[i];
	}

	/**
	 * Constructeur convertissant un tableau de double en polynome
	 * 
	 * @param tabfloat coef à recopier
	 */
	public Polynomes(double[] tabfloat) {
		super();
		degre = tabfloat.length - 1;
		monomes = tabfloat.clone();
	}

	// --------------------------------------------------------------------------------------------------------------

	/**
	 * On considère que deux nombres sont égaux si leur différence en valeur absolue
	 * est voisine de 0
	 * 
	 * @param a paramètre à tester
	 * @param b paramètre à tester
	 * @return vrai si a == b
	 */
	private boolean egalite(double a, double b) {
		final double EPSILON = 1E-6;
		return ((a - b) * (a - b) < EPSILON * EPSILON);
	}

	/**
	 * @param n degré du coefficient du polynome
	 * @return le coefficient de degré n, 0 si le coefficient n'existe pas
	 */
	private double coefficient(int n) {
		if (n >= monomes.length)
			return 0;
		else
			return monomes[n];
	}

	/**
	 * Compare deux polynomes.
	 * 
	 * @param other l'autre polynome
	 * @return {@code true} si le polynome {@code this} égale {@code other};
	 *         {@code false} autrement
	 */
	@Override
	public boolean equals(Object other) {
		if (other == this)
			return true;
		if (other == null)
			return false;
		if (other.getClass() != getClass())
			return false;
		Polynomes that = (Polynomes) other;
		if (degre != that.degre)
			return false;
		for (int i = degre; i >= 0; i--)
			if (coefficient(i) != that.coefficient(i))
				return false;
		return true;
	}

	/**
	 * Compare 2 polynomes.
	 *
	 * @param that l'autre polynome
	 * @return {@code 0} si les 2 polynomes sont égaux -1 si le polynome appelant
	 *         est inférieur à l'autre polynome +1 s'il est plus grand
	 */
	public int compareTo(Polynomes that) {
		if (degre < that.degre)
			return -1;
		if (degre > that.degre)
			return +1;
		for (int i = degre; i >= 0; i--) {
			if (this.monomes[i] < that.monomes[i])
				return -1;
			if (this.monomes[i] > that.monomes[i])
				return +1;
		}
		return 0;
	}

	/**
	 * @return true si le polynome est nul, false sinon
	 */
	public boolean isNul() {
		return degre == 0 && monomes[0] == 0;
	}
	
	/**
	 * Crée une chaine de caractères représentant un monome. Cette méthode est
	 * utilisée en particulier pour l'affichage
	 * 
	 * @param coef  du monome
	 * @param degre du monome
	 * @return la chaine représentant le monome
	 */
	private String toStringMonome(double coef, int deg) {
		if (deg == 0)
			return new Double(coef).toString();
		else if (deg == 1)
			return coef + "*X";
		else
			return coef + "*X^" + deg;
	}

	/**
	 * Crée un monome
	 * 
	 * @param coef  du monome créé
	 * @param degre du monome crée
	 * @return le monome
	 */
	private Polynomes creeMonome(double coef, int degre) {
		Polynomes monome = new Polynomes(degre);
		monome.monomes[degre] = coef;
		return monome;
	}

	/**
	 * Réduction du polynome
	 * 
	 * @return le polynome réduit
	 */
	private Polynomes reduction() {
		int i = degre;
		while (i > 0 && egalite(monomes[i], 0)) {
			i--;
			degre--;
		}
		return this;
	}

	/**
	 * Normalisation du polynome par rapport à son plus fort degré
	 * 
	 * @return le polynome normalisé
	 */
	private Polynomes normalise() {
		double coefDominant = monomes[degre];
		for (int i = 0; i <= degre; i++)
			monomes[i] /= coefDominant;
		return this;
	}

	// --------------------------------------------------------------------------------------------------------------

	/**
	 * Saisie d'un polynome au clavier
	 * 
	 * @return le polynome saisi
	 */
	public static Polynomes litPolynome() {
		Polynomes polyLu;
		System.out.print("Degré du polynome : ");
		Scanner entree = new Scanner(System.in);
		polyLu = new Polynomes(entree.nextInt());
		for (int i = polyLu.degre; i >= 0; i--) {
			System.out.print("Coefficient d'ordre " + i + " : ");
			polyLu.monomes[i] = entree.nextDouble();
		}
		// entree.close();
		return polyLu;
	}

	/**
	 * Affiche le polynome
	 */
	@Override
	public String toString() {
		boolean affichePlus = false;
		String poly = "";
		if (degre == -1 || (degre == 0 && egalite(monomes[0], 0)))
			return "0";
		else
			for (int i = degre; i >= 0; i--) {
				double coef = coefficient(i);
				if (!egalite(coef, 0))
					if (coef > 0 && affichePlus)
						poly = poly + "+" + toStringMonome(coef, i);
					else {
						poly += toStringMonome(coef, i);
						affichePlus = true;
					}
			}
		return poly;
	}

	/**
	 * Somme de deux polynomes <tt>(this + p)</tt>
	 * 
	 * @param p un polynome
	 * @return this + p
	 */
	public Polynomes sommePolynomes(Polynomes p) {
		Polynomes resultat = (degre > p.degre) ? new Polynomes(degre) : new Polynomes(p.degre);
		for (int i = 0; i <= resultat.degre; i++)
			resultat.monomes[i] = coefficient(i) + p.coefficient(i);
		return resultat.reduction();
	}

	/**
	 * Différence de deux polynomes <tt>(this - p)</tt>
	 * 
	 * @param p un polynome
	 * @return this - p
	 */
	public Polynomes soustractionPolynomes(Polynomes p) {
		Polynomes resultat = (degre > p.degre) ? new Polynomes(degre) : new Polynomes(p.degre);
		for (int i = 0; i <= resultat.degre; i++)
			resultat.monomes[i] = coefficient(i) - p.coefficient(i);
		return resultat.reduction();
	}

	/**
	 * Produit de deux polynomes <tt>(this * p)</tt>
	 * 
	 * @param p un polynome
	 * @return this * p
	 */

	public Polynomes produitPolynomes(Polynomes that) {
		Polynomes poly = new Polynomes(degre + that.degre);
		for (int i = 0; i <= degre; i++)
			for (int j = 0; j <= that.degre; j++)
				poly.monomes[i + j] += (coefficient(i) * that.coefficient(j));
		return poly.reduction();
	}

	/**
	 * Dérive le polynome
	 * 
	 * @return le polynome dérivé
	 */
	public Polynomes derivePolynomes() {
		Polynomes resultat = new Polynomes(degre - 1);
		for (int i = 1; i <= degre; i++)
			resultat.monomes[i - 1] = coefficient(i) * i;
		return resultat;
	}

	/**
	 * Division euclidienne du polynome par <tt>b</tt>.
	 * 
	 * Si <tt>this</tt> et <tt>b</tt> sont deux polynomes (avec <tt>b != 0</tt>), il
	 * existe un unique couple de polynomes <tt>q</tt> (le quotient) et <tt>r</tt>
	 * (le reste) tels que :
	 * <ul>
	 * <li>le degré de <tt>r</tt> est strictement inférieur à celui de <tt>b</tt>
	 * ;</li>
	 * <li><tt>this = bq + r</li>
	 * </ul>
	 * 
	 * @param b le diviseur.
	 * @return le reste de la division euclidienne.
	 */
	public Polynomes resteDivisionEuclidienne(Polynomes b) {
		// Attention b non nul
		// a = b * q + r
		// Initialisation
		// q = 0 donc r = a
		Polynomes q = new Polynomes();
		Polynomes r = new Polynomes(this);
		Polynomes m;
		while (r.degre >= b.degre) {
			m = creeMonome(r.coefficient(r.degre) / b.coefficient(b.degre), r.degre - b.degre);
			q = q.sommePolynomes(m);
			r = r.soustractionPolynomes(b.produitPolynomes(m));
		}
		return r;
	}

	/**
	 * Plus grand commun diviseur entre le polynome et <tt>p</tt> Application de
	 * l'algorithme d'Euclide
	 * 
	 * @param p un polynome
	 * @return le PGCD normalisé
	 */
	public Polynomes pgcdEuclide(Polynomes p) {
		Polynomes r1 = new Polynomes(this);
		Polynomes r2 = new Polynomes(p);
		Polynomes aux;
		while (r2.degre > 0) {
			aux = r2;
			r2 = r1.resteDivisionEuclidienne(r2);
			r1 = aux;
		}
		return r1.normalise();
	}

	
	/**
	 * Plus grand commun diviseur entre le polynome et <tt>p</tt> Application de
	 * l'algorithme d'Euclide. Version récursive.
	 * 
	 * @param p un polynome
	 * @return
	 */
	public Polynomes pgcdRecursifEuclide(Polynomes p) {
		Polynomes r = resteDivisionEuclidienne(p);
		if (r.isNul())
			return p.normalise();
			else return p.pgcdRecursifEuclide(r);		
	}

	
	// --------------------------------------------------------------------------------------------------------------

	public static void main(String args[]) {

		/*
		 * Polynomes p1 = litPolynome(); System.out.print("p1(x) = ");
		 * System.out.println(p1); Polynomes p2 = litPolynome();
		 * System.out.print("p2(x) = "); System.out.println(p2);
		 */
		Polynomes p1 = new Polynomes(new double[] { -2, -1, 1, 0, -2, 1 });
		Polynomes p2 = new Polynomes(new double[] { -2, -1, -1, 1 });
		System.out.print("p1(x) + p2(x) = ");
		System.out.println(p1.sommePolynomes(p2));
		System.out.print("p1(x) - p2(x) = ");
		System.out.println(p1.soustractionPolynomes(p2));
		System.out.print("p1(x) * p2(x) = ");
		System.out.println(p1.produitPolynomes(p2));
		System.out.print("p1'(x) = ");
		System.out.println(p1.derivePolynomes());
		System.out.print("p2'(x) = ");
		System.out.println(p2.derivePolynomes());
		System.out.print("p1(x) % p2(x) = ");
		System.out.println(p1.resteDivisionEuclidienne(p2));
		System.out.print("PGCD(p1,p2) = ");
		System.out.println(p1.pgcdEuclide(p2));
		System.out.print("PGCD(p1,p2) = ");
		System.out.println(p1.pgcdRecursifEuclide(p2));
		
	}

}
// --------------------------------------------------------------------------------------------------------------
